@@ -71,7 +71,16 @@ static void
 teardown_connection (TestConnection *test, gconstpointer data)
 {
   if (test->service)
-    g_object_unref (test->service);
+    {
+      g_socket_service_stop (test->service);
+      /* The outstanding accept_async will hold a ref on test->service,
+       * which we want to wait for it to release if we're valgrinding.
+       */
+      g_object_add_weak_pointer (G_OBJECT (test->service), (gpointer *)&test->service);
+      g_object_unref (test->service);
+      while (test->service)
+	g_main_context_iteration (NULL, TRUE);
+    }
 
   if (test->server_connection)
     {
@@ -316,6 +325,7 @@ test_verified_connection (TestConnection *test,
   test->client_connection = g_tls_client_connection_new (connection, test->identity, &error);
   g_assert_no_error (error);
   g_assert (test->client_connection);
+  g_object_unref (connection);
 
   g_tls_connection_set_database (G_TLS_CONNECTION (test->client_connection), test->database);
 
@@ -346,6 +356,7 @@ test_client_auth_connection (TestConnection *test,
   test->client_connection = g_tls_client_connection_new (connection, test->identity, &error);
   g_assert_no_error (error);
   g_assert (test->client_connection);
+  g_object_unref (connection);
 
   g_tls_connection_set_database (G_TLS_CONNECTION (test->client_connection), test->database);
 
@@ -384,6 +395,7 @@ test_connection_no_database (TestConnection *test,
   test->client_connection = g_tls_client_connection_new (connection, test->identity, &error);
   g_assert_no_error (error);
   g_assert (test->client_connection);
+  g_object_unref (connection);
 
   /* Overrides loading of the default database */
   g_tls_connection_set_database (G_TLS_CONNECTION (test->client_connection), NULL);
@@ -400,6 +412,93 @@ test_connection_no_database (TestConnection *test,
   g_main_loop_run (test->loop);
 }
 
+static void
+socket_client_connected (GObject      *source,
+			 GAsyncResult *result,
+			 gpointer      user_data)
+{
+  TestConnection *test = user_data;
+  GSocketConnection *connection;
+  GError *error = NULL;
+
+  connection = g_socket_client_connect_finish (G_SOCKET_CLIENT (source),
+					       result, &error);
+  g_assert_no_error (error);
+  test->client_connection = G_IO_STREAM (connection);
+
+  g_main_loop_quit (test->loop);
+}
+
+static void
+test_connection_socket_client (TestConnection *test,
+			       gconstpointer   data)
+{
+  GSocketClient *client;
+  GTlsCertificateFlags flags;
+  GSocketConnection *connection;
+  GIOStream *base;
+  GError *error = NULL;
+
+  start_server_service (test, G_TLS_AUTHENTICATION_NONE);
+  client = g_socket_client_new ();
+  g_socket_client_set_tls (client, TRUE);
+  flags = G_TLS_CERTIFICATE_VALIDATE_ALL & ~G_TLS_CERTIFICATE_UNKNOWN_CA;
+  /* test->address doesn't match the server's cert */
+  flags = flags & ~G_TLS_CERTIFICATE_BAD_IDENTITY;
+  g_socket_client_set_tls_validation_flags (client, flags);
+
+  g_socket_client_connect_async (client, G_SOCKET_CONNECTABLE (test->address),
+				 NULL, socket_client_connected, test);
+  g_main_loop_run (test->loop);
+
+  connection = (GSocketConnection *)test->client_connection;
+  test->client_connection = NULL;
+
+  g_assert (G_IS_TCP_WRAPPER_CONNECTION (connection));
+  base = g_tcp_wrapper_connection_get_base_io_stream (G_TCP_WRAPPER_CONNECTION (connection));
+  g_assert (G_IS_TLS_CONNECTION (base));
+
+  g_io_stream_close (G_IO_STREAM (connection), NULL, &error);
+  g_assert_no_error (error);
+  g_object_unref (connection);
+
+  g_object_unref (client);
+}
+
+static void
+socket_client_failed (GObject      *source,
+		      GAsyncResult *result,
+		      gpointer      user_data)
+{
+  TestConnection *test = user_data;
+  GError *error = NULL;
+
+  g_socket_client_connect_finish (G_SOCKET_CLIENT (source),
+				  result, &error);
+  g_assert_error (error, G_TLS_ERROR, G_TLS_ERROR_BAD_CERTIFICATE);
+  g_clear_error (&error);
+
+  g_main_loop_quit (test->loop);
+}
+
+static void
+test_connection_socket_client_failed (TestConnection *test,
+				      gconstpointer   data)
+{
+  GSocketClient *client;
+
+  start_server_service (test, G_TLS_AUTHENTICATION_NONE);
+  client = g_socket_client_new ();
+  g_socket_client_set_tls (client, TRUE);
+  /* this time we don't adjust the validation flags */
+
+  g_socket_client_connect_async (client, G_SOCKET_CONNECTABLE (test->address),
+				 NULL, socket_client_failed, test);
+  g_main_loop_run (test->loop);
+
+  g_object_unref (client);
+}
+
 /* -----------------------------------------------------------------------------
  * CERTIFICATE TESTS
  */
@@ -581,6 +680,11 @@ teardown_verify (TestVerify      *test,
 			     (gpointer *)&test->database);
   g_object_unref (test->database);
   g_assert (test->database == NULL);
+
+  g_object_add_weak_pointer (G_OBJECT (test->identity),
+			     (gpointer *)&test->identity);
+  g_object_unref (test->identity);
+  g_assert (test->identity == NULL);
 }
 
 static void
@@ -1010,6 +1114,10 @@ main (int   argc,
               setup_connection, test_client_auth_rehandshake, teardown_connection);
   g_test_add ("/tls/connection/no-database", TestConnection, NULL,
               setup_connection, test_connection_no_database, teardown_connection);
+  g_test_add ("/tls/connection/socket-client", TestConnection, NULL,
+              setup_connection, test_connection_socket_client, teardown_connection);
+  g_test_add ("/tls/connection/socket-client-failed", TestConnection, NULL,
+              setup_connection, test_connection_socket_client_failed, teardown_connection);
 
   g_test_add_func ("/tls/backend/default-database-is-singleton",
                    test_default_database_is_singleton);
@@ -1054,5 +1162,9 @@ main (int   argc,
   ret = g_test_run();
 
   g_free (source_dir);
+
+  /* for valgrinding */
+  g_main_context_unref (g_main_context_default ());
+
   return ret;
 }
